@@ -210,8 +210,7 @@ impl ScanExec {
 
         if iter.is_null() {
             return Err(CometError::from(ExecutionError::GeneralError(format!(
-                "Null batch iterator object. Plan id: {}",
-                exec_context_id
+                "Null batch iterator object. Plan id: {exec_context_id}"
             ))));
         }
 
@@ -303,14 +302,14 @@ fn scan_schema(input_batch: &InputBatch, data_types: &[DataType]) -> SchemaRef {
                 .map(|(idx, c)| {
                     let datatype = ScanExec::unpack_dictionary_type(c.data_type());
                     // We don't use the field name. Put a placeholder.
-                    Field::new(format!("col_{}", idx), datatype, true)
+                    Field::new(format!("col_{idx}"), datatype, true)
                 })
                 .collect::<Vec<Field>>()
         }
         _ => data_types
             .iter()
             .enumerate()
-            .map(|(idx, dt)| Field::new(format!("col_{}", idx), dt.clone(), true))
+            .map(|(idx, dt)| Field::new(format!("col_{idx}"), dt.clone(), true))
             .collect(),
     };
 
@@ -497,14 +496,14 @@ pub enum InputBatch {
     /// The end of input batches.
     EOF,
 
-    /// A normal batch with columns and number of rows.
-    /// It is possible to have zero-column batch with non-zero number of rows,
+    /// A normal batch with columns and a number of rows.
+    /// It is possible to have a zero-column batch with a non-zero number of rows,
     /// i.e. reading empty schema from scan.
     Batch(Vec<ArrayRef>, usize),
 }
 
 impl InputBatch {
-    /// Constructs a `InputBatch` from columns and optional number of rows.
+    /// Constructs an ` InputBatch ` from columns and an optional number of rows.
     /// If `num_rows` is none, this function will calculate it from given
     /// columns.
     pub fn new(columns: Vec<ArrayRef>, num_rows: Option<usize>) -> Self {
