@@ -17,9 +17,10 @@
 
 use crate::hash_funcs::*;
 use crate::{
-    spark_ceil, spark_date_add, spark_date_sub, spark_decimal_div, spark_decimal_integral_div,
-    spark_floor, spark_hex, spark_isnan, spark_make_decimal, spark_read_side_padding, spark_round,
-    spark_rpad, spark_unhex, spark_unscaled_value, SparkChrFunc,
+    spark_array_repeat, spark_ceil, spark_date_add, spark_date_sub, spark_decimal_div,
+    spark_decimal_integral_div, spark_floor, spark_hex, spark_isnan, spark_make_decimal,
+    spark_read_side_padding, spark_round, spark_rpad, spark_unhex, spark_unscaled_value,
+    SparkBitwiseCount, SparkBitwiseGet, SparkBitwiseNot, SparkChrFunc, SparkDateTrunc,
 };
 use arrow::datatypes::DataType;
 use datafusion::common::{DataFusionError, Result as DataFusionResult};
@@ -111,7 +112,6 @@ pub fn create_comet_physical_fun(
             let func = Arc::new(spark_xxhash64);
             make_comet_scalar_udf!("xxhash64", func, without data_type)
         }
-        "chr" => Ok(Arc::new(ScalarUDF::new_from_impl(SparkChrFunc::default()))),
         "isnan" => {
             let func = Arc::new(spark_isnan);
             make_comet_scalar_udf!("isnan", func, without data_type)
@@ -140,6 +140,10 @@ pub fn create_comet_physical_fun(
             let func = Arc::new(spark_date_sub);
             make_comet_scalar_udf!("date_sub", func, without data_type)
         }
+        "array_repeat" => {
+            let func = Arc::new(spark_array_repeat);
+            make_comet_scalar_udf!("array_repeat", func, without data_type)
+        }
         _ => registry.udf(fun_name).map_err(|e| {
             DataFusionError::Execution(format!(
                 "Function {fun_name} not found in the registry: {e}",
@@ -148,6 +152,26 @@ pub fn create_comet_physical_fun(
     }
 }
 
+fn all_scalar_functions() -> Vec<Arc<ScalarUDF>> {
+    vec![
+        Arc::new(ScalarUDF::new_from_impl(SparkChrFunc::default())),
+        Arc::new(ScalarUDF::new_from_impl(SparkBitwiseNot::default())),
+        Arc::new(ScalarUDF::new_from_impl(SparkBitwiseCount::default())),
+        Arc::new(ScalarUDF::new_from_impl(SparkBitwiseGet::default())),
+        Arc::new(ScalarUDF::new_from_impl(SparkDateTrunc::default())),
+    ]
+}
+
+/// Registers all custom UDFs
+pub fn register_all_comet_functions(registry: &mut dyn FunctionRegistry) -> DataFusionResult<()> {
+    // This will override existing UDFs with the same name
+    all_scalar_functions()
+        .into_iter()
+        .try_for_each(|udf| registry.register_udf(udf).map(|_| ()))?;
+
+    Ok(())
+}
+
 struct CometScalarFunction {
     name: String,
     signature: Signature,
