@@ -19,12 +19,19 @@
 
 package org.apache.comet
 
+import java.io.FileNotFoundException
+import java.lang.management.ManagementFactory
+
+import scala.util.matching.Regex
+
 import org.apache.spark._
 import org.apache.spark.internal.Logging
+import org.apache.spark.network.util.ByteUnit
 import org.apache.spark.sql.comet.CometMetricNode
 import org.apache.spark.sql.vectorized._
 
-import org.apache.comet.CometConf.{COMET_BATCH_SIZE, COMET_BLOCKING_THREADS, COMET_DEBUG_ENABLED, COMET_EXEC_MEMORY_POOL_TYPE, COMET_EXPLAIN_NATIVE_ENABLED, COMET_METRICS_UPDATE_INTERVAL, COMET_WORKER_THREADS}
+import org.apache.comet.CometConf.{COMET_BATCH_SIZE, COMET_DEBUG_ENABLED, COMET_EXEC_MEMORY_POOL_TYPE, COMET_EXPLAIN_NATIVE_ENABLED, COMET_METRICS_UPDATE_INTERVAL}
+import org.apache.comet.Tracing.withTrace
 import org.apache.comet.vector.NativeUtil
 
 /**
@@ -56,16 +63,27 @@ class CometExecIterator(
     extends Iterator[ColumnarBatch]
     with Logging {
 
+  private val tracingEnabled = CometConf.COMET_TRACING_ENABLED.get()
+  private val memoryMXBean = ManagementFactory.getMemoryMXBean
   private val nativeLib = new Native()
   private val nativeUtil = new NativeUtil()
+  private val cometTaskMemoryManager = new CometTaskMemoryManager(id)
   private val cometBatchIterators = inputs.map { iterator =>
     new CometBatchIterator(iterator, nativeUtil)
   }.toArray
   private val plan = {
     val conf = SparkEnv.get.conf
-    // Only enable unified memory manager when off-heap mode is enabled. Otherwise,
-    // we'll use the built-in memory pool from DF, and initializes with `memory_limit`
-    // and `memory_fraction` below.
+    val localDiskDirs = SparkEnv.get.blockManager.getLocalDiskDirs
+
+    val offHeapMode = CometSparkSessionExtensions.isOffHeapEnabled(conf)
+    val memoryLimit = if (offHeapMode) {
+      // in unified mode we share off-heap memory with Spark
+      ByteUnit.MiB.toBytes(conf.getSizeAsMb("spark.memory.offHeap.size"))
+    } else {
+      // we'll use the built-in memory pool from DF, and initializes with `memory_limit`
+      // and `memory_fraction` below.
+      CometSparkSessionExtensions.getCometMemoryOverhead(conf)
+    }
     nativeLib.createPlan(
       id,
       cometBatchIterators,
@@ -73,18 +91,17 @@ class CometExecIterator(
       numParts,
       nativeMetrics,
       metricsUpdateInterval = COMET_METRICS_UPDATE_INTERVAL.get(),
-      new CometTaskMemoryManager(id),
+      cometTaskMemoryManager,
+      localDiskDirs,
       batchSize = COMET_BATCH_SIZE.get(),
-      use_unified_memory_manager =
-        CometSparkSessionExtensions.cometUnifiedMemoryManagerEnabled(conf),
-      memory_pool_type = COMET_EXEC_MEMORY_POOL_TYPE.get(),
-      memory_limit = CometSparkSessionExtensions.getCometMemoryOverhead(conf),
-      memory_limit_per_task = getMemoryLimitPerTask(conf),
-      task_attempt_id = TaskContext.get().taskAttemptId,
+      offHeapMode,
+      memoryPoolType = COMET_EXEC_MEMORY_POOL_TYPE.get(),
+      memoryLimit,
+      memoryLimitPerTask = getMemoryLimitPerTask(conf),
+      taskAttemptId = TaskContext.get().taskAttemptId,
       debug = COMET_DEBUG_ENABLED.get(),
       explain = COMET_EXPLAIN_NATIVE_ENABLED.get(),
-      workerThreads = COMET_WORKER_THREADS.get(),
-      blockingThreads = COMET_BLOCKING_THREADS.get())
+      tracingEnabled)
   }
 
   private var nextBatch: Option[ColumnarBatch] = None
@@ -119,15 +136,53 @@ class CometExecIterator(
     }
   }
 
-  def getNextBatch(): Option[ColumnarBatch] = {
+  private def getNextBatch: Option[ColumnarBatch] = {
     assert(partitionIndex >= 0 && partitionIndex < numParts)
 
-    nativeUtil.getNextBatch(
-      numOutputCols,
-      (arrayAddrs, schemaAddrs) => {
-        val ctx = TaskContext.get()
-        nativeLib.executePlan(ctx.stageId(), partitionIndex, plan, arrayAddrs, schemaAddrs)
-      })
+    if (tracingEnabled) {
+      traceMemoryUsage()
+    }
+
+    val ctx = TaskContext.get()
+
+    try {
+      withTrace(
+        s"getNextBatch[JVM] stage=${ctx.stageId()}",
+        tracingEnabled, {
+          nativeUtil.getNextBatch(
+            numOutputCols,
+            (arrayAddrs, schemaAddrs) => {
+              nativeLib.executePlan(ctx.stageId(), partitionIndex, plan, arrayAddrs, schemaAddrs)
+            })
+        })
+    } catch {
+      case e: CometNativeException =>
+        val fileNotFoundPattern: Regex =
+          ("""^External: Object at location (.+?) not found: No such file or directory """ +
+            """\(os error \d+\)$""").r
+        val parquetError: Regex =
+          """^Parquet error: (?:.*)$""".r
+        e.getMessage match {
+          case fileNotFoundPattern(filePath) =>
+            // See org.apache.spark.sql.errors.QueryExecutionErrors.readCurrentFileNotFoundError
+            throw new SparkException(
+              errorClass = "_LEGACY_ERROR_TEMP_2055",
+              messageParameters = Map("message" -> e.getMessage),
+              cause = new FileNotFoundException(filePath)
+            ) // Can't use SparkFileNotFoundException because it's private.
+          case parquetError() =>
+            // See org.apache.spark.sql.errors.QueryExecutionErrors.failedToReadDataError
+            // See org.apache.parquet.hadoop.ParquetFileReader for error message.
+            throw new SparkException(
+              errorClass = "_LEGACY_ERROR_TEMP_2254",
+              messageParameters = Map("message" -> e.getMessage),
+              cause = new SparkException("File is not a Parquet file.", e))
+          case _ =>
+            throw e
+        }
+      case e: Throwable =>
+        throw e
+    }
   }
 
   override def hasNext: Boolean = {
@@ -145,7 +200,7 @@ class CometExecIterator(
       prevBatch = null
     }
 
-    nextBatch = getNextBatch()
+    nextBatch = getNextBatch
 
     if (nextBatch.isEmpty) {
       close()
@@ -181,6 +236,10 @@ class CometExecIterator(
       nativeUtil.close()
       nativeLib.releasePlan(plan)
 
+      if (tracingEnabled) {
+        traceMemoryUsage()
+      }
+
       // The allocator thoughts the exported ArrowArray and ArrowSchema structs are not released,
       // so it will report:
       // Caused by: java.lang.IllegalStateException: Memory was leaked by query.
@@ -203,4 +262,14 @@ class CometExecIterator(
       closed = true
     }
   }
+
+  private def traceMemoryUsage(): Unit = {
+    nativeLib.logMemoryUsage("jvm_heapUsed", memoryMXBean.getHeapMemoryUsage.getUsed)
+    val totalTaskMemory = cometTaskMemoryManager.internal.getMemoryConsumptionForThisTask
+    val cometTaskMemory = cometTaskMemoryManager.getUsed
+    val sparkTaskMemory = totalTaskMemory - cometTaskMemory
+    val threadId = Thread.currentThread().getId
+    nativeLib.logMemoryUsage(s"task_memory_comet_$threadId", cometTaskMemory)
+    nativeLib.logMemoryUsage(s"task_memory_spark_$threadId", sparkTaskMemory)
+  }
 }
