@@ -27,12 +27,12 @@ import scala.concurrent.duration.NANOSECONDS
 import scala.util.control.NonFatal
 
 import org.apache.spark.{broadcast, Partition, SparkContext, TaskContext}
-import org.apache.spark.comet.shims.ShimCometBroadcastExchangeExec
-import org.apache.spark.launcher.SparkLauncher
 import org.apache.spark.rdd.RDD
 import org.apache.spark.sql.catalyst.InternalRow
 import org.apache.spark.sql.catalyst.expressions.Attribute
 import org.apache.spark.sql.catalyst.plans.logical.Statistics
+import org.apache.spark.sql.catalyst.plans.physical.{BroadcastMode, BroadcastPartitioning, Partitioning}
+import org.apache.spark.sql.comet.util.Utils
 import org.apache.spark.sql.errors.QueryExecutionErrors
 import org.apache.spark.sql.execution.{ColumnarToRowExec, SparkPlan, SQLExecution}
 import org.apache.spark.sql.execution.adaptive.{AQEShuffleReadExec, ShuffleQueryStageExec}
@@ -46,6 +46,7 @@ import org.apache.spark.util.io.ChunkedByteBuffer
 import com.google.common.base.Objects
 
 import org.apache.comet.CometRuntimeException
+import org.apache.comet.shims.ShimCometBroadcastExchangeExec
 
 /**
  * A [[CometBroadcastExchangeExec]] collects, transforms and finally broadcasts the result of a
@@ -57,17 +58,15 @@ import org.apache.comet.CometRuntimeException
  * Note that this class cannot extend `CometExec` as usual similar to other Comet operators. As
  * the trait `BroadcastExchangeLike` in Spark extends abstract class `Exchange`, it limits the
  * flexibility to extend `CometExec` and `Exchange` at the same time.
- *
- * Note that this only supports Spark 3.4 and later, because the serialization class
- * `ChunkedByteBuffer` is only serializable in Spark 3.4 and later.
  */
 case class CometBroadcastExchangeExec(
     originalPlan: SparkPlan,
     override val output: Seq[Attribute],
+    mode: BroadcastMode,
     override val child: SparkPlan)
     extends BroadcastExchangeLike
-    with ShimCometBroadcastExchangeExec {
-  import CometBroadcastExchangeExec._
+    with ShimCometBroadcastExchangeExec
+    with CometPlan {
 
   override val runId: UUID = UUID.randomUUID
 
@@ -79,7 +78,7 @@ case class CometBroadcastExchangeExec(
     "broadcastTime" -> SQLMetrics.createTimingMetric(sparkContext, "time to broadcast"))
 
   override def doCanonicalize(): SparkPlan = {
-    CometBroadcastExchangeExec(null, null, child.canonicalized)
+    CometBroadcastExchangeExec(null, null, mode, child.canonicalized)
   }
 
   override def runtimeStatistics: Statistics = {
@@ -88,6 +87,8 @@ case class CometBroadcastExchangeExec(
     Statistics(dataSize, Some(rowCount))
   }
 
+  override def outputPartitioning: Partitioning = BroadcastPartitioning(mode)
+
   @transient
   private lazy val promise = Promise[broadcast.Broadcast[Any]]()
 
@@ -117,11 +118,7 @@ case class CometBroadcastExchangeExec(
       session,
       CometBroadcastExchangeExec.executionContext) {
       try {
-        // Setup a job group here so later it may get cancelled by groupId if necessary.
-        sparkContext.setJobGroup(
-          runId.toString,
-          s"broadcast exchange (runId $runId)",
-          interruptOnCancel = true)
+        setJobGroupOrTag(sparkContext, this)
         val beforeCollect = System.nanoTime()
 
         val countsAndBytes = child match {
@@ -167,9 +164,10 @@ case class CometBroadcastExchangeExec(
         val dataSize = batches.map(_.size).sum
 
         longMetric("dataSize") += dataSize
-        if (dataSize >= MAX_BROADCAST_TABLE_BYTES) {
+        val maxBytes = maxBroadcastTableBytes(conf)
+        if (dataSize >= maxBytes) {
           throw QueryExecutionErrors.cannotBroadcastTableOverMaxTableBytesError(
-            MAX_BROADCAST_TABLE_BYTES,
+            maxBytes,
             dataSize)
         }
 
@@ -177,7 +175,8 @@ case class CometBroadcastExchangeExec(
         longMetric("buildTime") += NANOSECONDS.toMillis(beforeBroadcast - beforeBuild)
 
         // (3.4 only) SPARK-39983 - Broadcast the relation without caching the unserialized object.
-        val broadcasted = doBroadcast(sparkContext, batches)
+        val broadcasted = sparkContext
+          .broadcastInternal(batches, true)
           .asInstanceOf[broadcast.Broadcast[Any]]
         longMetric("broadcastTime") += NANOSECONDS.toMillis(System.nanoTime() - beforeBroadcast)
         val executionId = sparkContext.getLocalProperty(SQLExecution.EXECUTION_ID_KEY)
@@ -189,16 +188,7 @@ case class CometBroadcastExchangeExec(
         // SparkFatalException, which is a subclass of Exception. ThreadUtils.awaitResult
         // will catch this exception and re-throw the wrapped fatal throwable.
         case oe: OutOfMemoryError =>
-          // Spark 3.4 has two parameters for `notEnoughMemoryToBuildAndBroadcastTableError`, which
-          // is different to Spark 3.3. We simply create the error message here.
-          val error =
-            new OutOfMemoryError(
-              "Not enough memory to build and broadcast the table to all " +
-                "worker nodes. As a workaround, you can either disable broadcast by setting " +
-                s"${SQLConf.AUTO_BROADCASTJOIN_THRESHOLD.key} to -1 or increase the spark " +
-                s"driver memory by setting ${SparkLauncher.DRIVER_MEMORY} to a higher value.")
-              .initCause(oe.getCause)
-          val ex = new SparkFatalException(error)
+          val ex = new SparkFatalException(oe)
           promise.tryFailure(ex)
           throw ex
         case e if !NonFatal(e) =>
@@ -241,7 +231,7 @@ case class CometBroadcastExchangeExec(
       case ex: TimeoutException =>
         logError(s"Could not execute broadcast in $timeout secs.", ex)
         if (!relationFuture.isDone) {
-          sparkContext.cancelJobGroup(runId.toString)
+          cancelJobGroup(sparkContext, this)
           relationFuture.cancel(true)
         }
         throw QueryExecutionErrors.executeBroadcastTimeoutError(timeout, Some(ex))
@@ -267,8 +257,6 @@ case class CometBroadcastExchangeExec(
 }
 
 object CometBroadcastExchangeExec {
-  val MAX_BROADCAST_TABLE_BYTES: Long = 8L << 30
-
   private[comet] val executionContext = ExecutionContext.fromExecutorService(
     ThreadUtils.newDaemonCachedThreadPool(
       "comet-broadcast-exchange",
@@ -299,7 +287,7 @@ class CometBatchRDD(
   override def compute(split: Partition, context: TaskContext): Iterator[ColumnarBatch] = {
     val partition = split.asInstanceOf[CometBatchPartition]
     partition.value.value.toIterator
-      .flatMap(CometExec.decodeBatches(_, this.getClass.getSimpleName))
+      .flatMap(Utils.decodeBatches(_, this.getClass.getSimpleName))
   }
 }
 
