@@ -33,6 +33,7 @@ import org.apache.spark.sql.catalyst.plans.QueryPlan
 import org.apache.spark.sql.catalyst.plans.physical._
 import org.apache.spark.sql.catalyst.util.CaseInsensitiveMap
 import org.apache.spark.sql.comet.shims.ShimCometScanExec
+import org.apache.spark.sql.errors.QueryExecutionErrors
 import org.apache.spark.sql.execution._
 import org.apache.spark.sql.execution.datasources._
 import org.apache.spark.sql.execution.datasources.parquet.{ParquetFileFormat, ParquetOptions}
@@ -43,14 +44,22 @@ import org.apache.spark.sql.vectorized.ColumnarBatch
 import org.apache.spark.util.SerializableConfiguration
 import org.apache.spark.util.collection._
 
-import org.apache.comet.{CometConf, DataTypeSupport, MetricsSupport}
+import org.apache.comet.{CometConf, MetricsSupport}
 import org.apache.comet.parquet.{CometParquetFileFormat, CometParquetPartitionReaderFactory}
 
 /**
  * Comet physical scan node for DataSource V1. Most of the code here follow Spark's
- * [[FileSourceScanExec]],
+ * [[FileSourceScanExec]].
+ *
+ * This is a hybrid scan where the native plan will contain a `ScanExec` that reads batches of
+ * data from the JVM via JNI. The ultimate source of data may be a JVM implementation such as
+ * Spark readers, or could be the `native_comet` or `native_iceberg_compat` native scans.
+ *
+ * Note that scanImpl can only be `native_datafusion` after CometScanRule runs and before
+ * CometExecRule runs. It will never be set to `native_datafusion` at execution time
  */
 case class CometScanExec(
+    scanImpl: String,
     @transient relation: HadoopFsRelation,
     output: Seq[Attribute],
     requiredSchema: StructType,
@@ -65,6 +74,8 @@ case class CometScanExec(
     with ShimCometScanExec
     with CometPlan {
 
+  assert(scanImpl != CometConf.SCAN_AUTO)
+
   // FIXME: ideally we should reuse wrapped.supportsColumnar, however that fails many tests
   override lazy val supportsColumnar: Boolean =
     relation.fileFormat.supportBatch(relation.sparkSession, schema)
@@ -307,7 +318,7 @@ case class CometScanExec(
         .groupBy { f =>
           BucketingUtils
             .getBucketId(new Path(f.filePath.toString()).getName)
-            .getOrElse(throw invalidBucketFile(f.filePath.toString(), sparkContext.version))
+            .getOrElse(throw QueryExecutionErrors.invalidBucketFile(f.filePath.toString()))
         }
 
     val prunedFilesGroupedToBuckets = if (optionalBucketSet.isDefined) {
@@ -413,16 +424,8 @@ case class CometScanExec(
       readFile: (PartitionedFile) => Iterator[InternalRow],
       partitions: Seq[FilePartition]): RDD[InternalRow] = {
     val hadoopConf = relation.sparkSession.sessionState.newHadoopConfWithOptions(relation.options)
-    val usingDataFusionReader: Boolean = {
-      hadoopConf.getBoolean(
-        CometConf.COMET_NATIVE_SCAN_ENABLED.key,
-        CometConf.COMET_NATIVE_SCAN_ENABLED.defaultValue.get) &&
-      hadoopConf
-        .get(
-          CometConf.COMET_NATIVE_SCAN_IMPL.key,
-          CometConf.COMET_NATIVE_SCAN_IMPL.defaultValueString)
-        .equalsIgnoreCase(CometConf.SCAN_NATIVE_ICEBERG_COMPAT)
-    }
+    val usingDataFusionReader: Boolean = scanImpl == CometConf.SCAN_NATIVE_ICEBERG_COMPAT
+
     val prefetchEnabled = hadoopConf.getBoolean(
       CometConf.COMET_SCAN_PREFETCH_ENABLED.key,
       CometConf.COMET_SCAN_PREFETCH_ENABLED.defaultValue.get) &&
@@ -434,6 +437,7 @@ case class CometScanExec(
       val broadcastedConf =
         fsRelation.sparkSession.sparkContext.broadcast(new SerializableConfiguration(hadoopConf))
       val partitionReaderFactory = CometParquetPartitionReaderFactory(
+        scanImpl == CometConf.SCAN_NATIVE_ICEBERG_COMPAT,
         sqlConf,
         broadcastedConf,
         requiredSchema,
@@ -458,20 +462,14 @@ case class CometScanExec(
     }
   }
 
-  // Filters unused DynamicPruningExpression expressions - one which has been replaced
-  // with DynamicPruningExpression(Literal.TrueLiteral) during Physical Planning
-  private def filterUnusedDynamicPruningExpressions(
-      predicates: Seq[Expression]): Seq[Expression] = {
-    predicates.filterNot(_ == DynamicPruningExpression(Literal.TrueLiteral))
-  }
-
   override def doCanonicalize(): CometScanExec = {
     CometScanExec(
+      scanImpl,
       relation,
       output.map(QueryPlan.normalizeExpressions(_, output)),
       requiredSchema,
       QueryPlan.normalizePredicates(
-        filterUnusedDynamicPruningExpressions(partitionFilters),
+        CometScanUtils.filterUnusedDynamicPruningExpressions(partitionFilters),
         output),
       optionalBucketSet,
       optionalNumCoalescedBuckets,
@@ -482,21 +480,12 @@ case class CometScanExec(
   }
 }
 
-object CometScanExec extends DataTypeSupport {
+object CometScanExec {
 
-  override def isAdditionallySupported(dt: DataType): Boolean = {
-    if (CometConf.COMET_NATIVE_SCAN_IMPL.get() == CometConf.SCAN_NATIVE_ICEBERG_COMPAT) {
-      // TODO add array and map
-      dt match {
-        case s: StructType => s.fields.map(_.dataType).forall(isTypeSupported)
-        case _ => false
-      }
-    } else {
-      false
-    }
-  }
-
-  def apply(scanExec: FileSourceScanExec, session: SparkSession): CometScanExec = {
+  def apply(
+      scanExec: FileSourceScanExec,
+      session: SparkSession,
+      scanImpl: String): CometScanExec = {
     // TreeNode.mapProductIterator is protected method.
     def mapProductIterator[B: ClassTag](product: Product, f: Any => B): Array[B] = {
       val arr = Array.ofDim[B](product.productArity)
@@ -514,13 +503,14 @@ object CometScanExec extends DataTypeSupport {
     // https://github.com/apache/arrow-datafusion-comet/issues/190
     def transform(arg: Any): AnyRef = arg match {
       case _: HadoopFsRelation =>
-        scanExec.relation.copy(fileFormat = new CometParquetFileFormat)(session)
+        scanExec.relation.copy(fileFormat = new CometParquetFileFormat(scanImpl))(session)
       case other: AnyRef => other
       case null => null
     }
-    val newArgs = mapProductIterator(scanExec, transform(_))
+    val newArgs = mapProductIterator(scanExec, transform)
     val wrapped = scanExec.makeCopy(newArgs).asInstanceOf[FileSourceScanExec]
     val batchScanExec = CometScanExec(
+      scanImpl,
       wrapped.relation,
       wrapped.output,
       wrapped.requiredSchema,
@@ -540,4 +530,5 @@ object CometScanExec extends DataTypeSupport {
     // of ParquetFileFormat.
     fileFormat.getClass().equals(classOf[ParquetFileFormat])
   }
+
 }
