@@ -36,10 +36,10 @@ import org.apache.spark.sql.catalyst.expressions.aggregate.{AggregateMode, Bloom
 import org.apache.spark.sql.comet._
 import org.apache.spark.sql.comet.execution.shuffle.{CometColumnarShuffle, CometShuffleExchangeExec}
 import org.apache.spark.sql.execution.{CollectLimitExec, ProjectExec, SQLExecution, UnionExec}
-import org.apache.spark.sql.execution.adaptive.AdaptiveSparkPlanExec
+import org.apache.spark.sql.execution.adaptive.{AdaptiveSparkPlanExec, BroadcastQueryStageExec}
 import org.apache.spark.sql.execution.datasources.parquet.ParquetFileFormat
 import org.apache.spark.sql.execution.exchange.{BroadcastExchangeExec, ReusedExchangeExec, ShuffleExchangeExec}
-import org.apache.spark.sql.execution.joins.{BroadcastNestedLoopJoinExec, CartesianProductExec, SortMergeJoinExec}
+import org.apache.spark.sql.execution.joins.{BroadcastHashJoinExec, BroadcastNestedLoopJoinExec, CartesianProductExec, SortMergeJoinExec}
 import org.apache.spark.sql.execution.reuse.ReuseExchangeAndSubquery
 import org.apache.spark.sql.execution.window.WindowExec
 import org.apache.spark.sql.expressions.Window
@@ -49,7 +49,7 @@ import org.apache.spark.sql.internal.SQLConf.SESSION_LOCAL_TIMEZONE
 import org.apache.spark.unsafe.types.UTF8String
 
 import org.apache.comet.{CometConf, ExtendedExplainInfo}
-import org.apache.comet.CometSparkSessionExtensions.{isSpark33Plus, isSpark34Plus, isSpark35Plus, isSpark40Plus}
+import org.apache.comet.CometSparkSessionExtensions.{isSpark35Plus, isSpark40Plus}
 import org.apache.comet.testing.{DataGenOptions, ParquetGenerator}
 
 class CometExecSuite extends CometTestBase {
@@ -59,7 +59,9 @@ class CometExecSuite extends CometTestBase {
   override protected def test(testName: String, testTags: Tag*)(testFun: => Any)(implicit
       pos: Position): Unit = {
     super.test(testName, testTags: _*) {
-      withSQLConf(CometConf.COMET_EXEC_SHUFFLE_ENABLED.key -> "true") {
+      withSQLConf(
+        CometConf.COMET_EXEC_SHUFFLE_ENABLED.key -> "true",
+        CometConf.COMET_NATIVE_SCAN_IMPL.key -> CometConf.SCAN_AUTO) {
         testFun
       }
     }
@@ -81,7 +83,7 @@ class CometExecSuite extends CometTestBase {
           .create()
 
         val df = sql("SELECT * FROM test_data ORDER BY c1 LIMIT 3")
-        checkSparkAnswer(df)
+        checkSparkAnswerAndOperator(df)
       }
     }
   }
@@ -115,7 +117,7 @@ class CometExecSuite extends CometTestBase {
               "select * from dpp_fact join dpp_dim on fact_date = dim_date where dim_id > 7")
           val (_, cometPlan) = checkSparkAnswer(df)
           val infos = new ExtendedExplainInfo().generateExtendedInfo(cometPlan)
-          assert(infos.contains("DPP not supported"))
+          assert(infos.contains("Dynamic Partition Pruning is not supported"))
         }
       }
     }
@@ -139,6 +141,32 @@ class CometExecSuite extends CometTestBase {
     }
   }
 
+  // repro for https://github.com/apache/datafusion-comet/issues/1251
+  test("subquery/exists-subquery/exists-orderby-limit.sql") {
+    withSQLConf(CometConf.COMET_SHUFFLE_MODE.key -> "jvm") {
+      val table = "src"
+      withTable(table) {
+        sql(s"CREATE TABLE $table (key INT, value STRING) USING PARQUET")
+        sql(s"INSERT INTO $table VALUES(238, 'val_238')")
+
+        // the subquery returns the distinct group by values
+        checkSparkAnswerAndOperator(s"""SELECT * FROM $table
+             |WHERE EXISTS (SELECT MAX(key)
+             |FROM $table
+             |GROUP BY value
+             |LIMIT 1
+             |OFFSET 2)""".stripMargin)
+
+        checkSparkAnswerAndOperator(s"""SELECT * FROM $table
+             |WHERE NOT EXISTS (SELECT MAX(key)
+             |FROM $table
+             |GROUP BY value
+             |LIMIT 1
+             |OFFSET 2)""".stripMargin)
+      }
+    }
+  }
+
   test("Sort on single struct should fallback to Spark") {
     withSQLConf(
       SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> "true",
@@ -229,7 +257,7 @@ class CometExecSuite extends CometTestBase {
           |)
           |SELECT *, (SELECT COUNT(*) FROM t2) FROM t2 LIMIT 10
           |""".stripMargin)
-      checkSparkAnswer(df)
+      checkSparkAnswerAndOperator(df)
     }
   }
 
@@ -260,7 +288,6 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("fix CometNativeExec.doCanonicalize for ReusedExchangeExec") {
-    assume(isSpark34Plus, "ChunkedByteBuffer is not serializable before Spark 3.4+")
     withSQLConf(
       CometConf.COMET_EXEC_BROADCAST_FORCE_ENABLED.key -> "true",
       SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> "false") {
@@ -290,7 +317,6 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("ReusedExchangeExec should work on CometBroadcastExchangeExec") {
-    assume(isSpark34Plus, "ChunkedByteBuffer is not serializable before Spark 3.4+")
     withSQLConf(
       CometConf.COMET_EXEC_BROADCAST_FORCE_ENABLED.key -> "true",
       SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> "false",
@@ -370,12 +396,10 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("Repeated shuffle exchange don't fail") {
-    assume(isSpark33Plus)
     Seq("true", "false").foreach { aqeEnabled =>
       withSQLConf(
         SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> aqeEnabled,
-        // `REQUIRE_ALL_CLUSTER_KEYS_FOR_DISTRIBUTION` is a new config in Spark 3.3+.
-        "spark.sql.requireAllClusterKeysForDistribution" -> "true",
+        SQLConf.REQUIRE_ALL_CLUSTER_KEYS_FOR_DISTRIBUTION.key -> "true",
         CometConf.COMET_SHUFFLE_MODE.key -> "jvm") {
         val df =
           Seq(("a", 1, 1), ("a", 2, 2), ("b", 1, 3), ("b", 1, 4)).toDF("key1", "key2", "value")
@@ -392,7 +416,6 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("try_sum should return null if overflow happens before merging") {
-    assume(isSpark33Plus, "try_sum is available in Spark 3.3+")
     val longDf = Seq(Long.MaxValue, Long.MaxValue, 2).toDF("v")
     val yearMonthDf = Seq(Int.MaxValue, Int.MaxValue, 2)
       .map(Period.ofMonths)
@@ -419,7 +442,6 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("CometBroadcastExchangeExec") {
-    assume(isSpark34Plus, "ChunkedByteBuffer is not serializable before Spark 3.4+")
     withSQLConf(CometConf.COMET_EXEC_BROADCAST_FORCE_ENABLED.key -> "true") {
       withParquetTable((0 until 5).map(i => (i, i + 1)), "tbl_a") {
         withParquetTable((0 until 5).map(i => (i, i + 1)), "tbl_b") {
@@ -445,7 +467,6 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("CometBroadcastExchangeExec: empty broadcast") {
-    assume(isSpark34Plus, "ChunkedByteBuffer is not serializable before Spark 3.4+")
     withSQLConf(CometConf.COMET_EXEC_BROADCAST_FORCE_ENABLED.key -> "true") {
       withParquetTable((0 until 5).map(i => (i, i + 1)), "tbl_a") {
         withParquetTable((0 until 5).map(i => (i, i + 1)), "tbl_b") {
@@ -525,13 +546,15 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("Comet native metrics: scan") {
-    // https://github.com/apache/datafusion-comet/issues/1441
-    assume(CometConf.COMET_NATIVE_SCAN_IMPL.get() != CometConf.SCAN_NATIVE_ICEBERG_COMPAT)
-
-    withSQLConf(CometConf.COMET_EXEC_ENABLED.key -> "true") {
+    withSQLConf(
+      CometConf.COMET_EXEC_ENABLED.key -> "true",
+      // TODO: update this test to work with native_iceberg_compat/auto,
+      // scan is set to native_comet for now as a workaround
+      // https://github.com/apache/datafusion-comet/issues/1882
+      CometConf.COMET_NATIVE_SCAN_IMPL.key -> CometConf.SCAN_NATIVE_COMET) {
       withTempDir { dir =>
         val path = new Path(dir.toURI.toString, "native-scan.parquet")
-        makeParquetFileAllTypes(path, dictionaryEnabled = true, 10000)
+        makeParquetFileAllPrimitiveTypes(path, dictionaryEnabled = true, 10000)
         withParquetTable(path.toString, "tbl") {
           val df = sql("SELECT * FROM tbl WHERE _2 > _3")
           df.collect()
@@ -663,7 +686,6 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("Comet native metrics: BroadcastHashJoin") {
-    assume(isSpark34Plus, "ChunkedByteBuffer is not serializable before Spark 3.4+")
     withParquetTable((0 until 5).map(i => (i, i + 1)), "t1") {
       withParquetTable((0 until 5).map(i => (i, i + 1)), "t2") {
         val df = sql("SELECT /*+ BROADCAST(t1) */ * FROM t1 INNER JOIN t2 ON t1._1 = t2._1")
@@ -748,7 +770,7 @@ class CometExecSuite extends CometTestBase {
              |""".stripMargin
 
         val df = sql(query)
-        checkSparkAnswer(df)
+        checkSparkAnswerAndOperator(df)
         val exchanges = stripAQEPlan(df.queryExecution.executedPlan).collect {
           case s: CometShuffleExchangeExec if s.shuffleType == CometColumnarShuffle =>
             s
@@ -758,6 +780,99 @@ class CometExecSuite extends CometTestBase {
     }
   }
 
+  test("Comet Shuffled Join should be optimized to CometBroadcastHashJoin by AQE") {
+    withSQLConf(
+      SQLConf.AUTO_BROADCASTJOIN_THRESHOLD.key -> "-1",
+      SQLConf.ADAPTIVE_AUTO_BROADCASTJOIN_THRESHOLD.key -> "10485760",
+      CometConf.COMET_EXEC_SHUFFLE_ENABLED.key -> "true",
+      CometConf.COMET_SHUFFLE_MODE.key -> "native") {
+      withParquetTable((0 until 100).map(i => (i, i + 1)), "tbl_a") {
+        withParquetTable((0 until 100).map(i => (i, i + 2)), "tbl_b") {
+          withParquetTable((0 until 100).map(i => (i, i + 3)), "tbl_c") {
+            val df = sql("""SELECT /*+ BROADCAST(c) */ a1, sum_b2, c._2 FROM (
+                |  SELECT a._1 a1, SUM(b._2) sum_b2 FROM tbl_a a
+                |  JOIN tbl_b b ON a._1 = b._1
+                |  GROUP BY a._1) t
+                |JOIN tbl_c c ON t.a1 = c._1
+                |""".stripMargin)
+            checkSparkAnswerAndOperator(df)
+
+            // Before AQE: 1 broadcast join
+            var broadcastHashJoinExec = stripAQEPlan(df.queryExecution.executedPlan).collect {
+              case s: CometBroadcastHashJoinExec => s
+            }
+            assert(broadcastHashJoinExec.length == 1)
+
+            // After AQE: shuffled join optimized to broadcast join
+            df.collect()
+            broadcastHashJoinExec = stripAQEPlan(df.queryExecution.executedPlan).collect {
+              case s: CometBroadcastHashJoinExec => s
+            }
+            assert(broadcastHashJoinExec.length == 2)
+          }
+        }
+      }
+    }
+  }
+
+  test("CometBroadcastExchange could be converted to rows using CometColumnarToRow") {
+    withSQLConf(
+      SQLConf.AUTO_BROADCASTJOIN_THRESHOLD.key -> "-1",
+      SQLConf.ADAPTIVE_AUTO_BROADCASTJOIN_THRESHOLD.key -> "10485760",
+      CometConf.COMET_EXEC_SHUFFLE_ENABLED.key -> "true",
+      CometConf.COMET_SHUFFLE_MODE.key -> "auto") {
+      withParquetTable((0 until 100).map(i => (i, i + 1)), "tbl_a") {
+        withParquetTable((0 until 100).map(i => (i, i + 2)), "tbl_b") {
+          withParquetTable((0 until 100).map(i => (i, i + 3)), "tbl_c") {
+            val df = sql("""SELECT /*+ BROADCAST(c) */ a1, sum_b2, c._2 FROM (
+                |  SELECT a._1 a1, SUM(b._2) sum_b2 FROM tbl_a a
+                |  JOIN tbl_b b ON a._1 = b._1
+                |  GROUP BY a._1) t
+                |JOIN tbl_c c ON t.a1 = c._1
+                |""".stripMargin)
+            checkSparkAnswerAndOperator(df)
+
+            // Before AQE: one CometBroadcastExchange, no CometColumnarToRow
+            var columnarToRowExec = stripAQEPlan(df.queryExecution.executedPlan).collect {
+              case s: CometColumnarToRowExec => s
+            }
+            assert(columnarToRowExec.isEmpty)
+
+            // Disable CometExecRule after the initial plan is generated. The CometSortMergeJoin and
+            // CometBroadcastHashJoin nodes in the initial plan will be converted to Spark BroadcastHashJoin
+            // during AQE. This will make CometBroadcastExchangeExec being converted to rows to be used by
+            // Spark BroadcastHashJoin.
+            withSQLConf(CometConf.COMET_EXEC_ENABLED.key -> "false") {
+              df.collect()
+            }
+
+            // After AQE: CometBroadcastExchange has to be converted to rows to conform to Spark
+            // BroadcastHashJoin.
+            val plan = stripAQEPlan(df.queryExecution.executedPlan)
+            columnarToRowExec = plan.collect { case s: CometColumnarToRowExec =>
+              s
+            }
+            assert(columnarToRowExec.length == 1)
+
+            // This ColumnarToRowExec should be the immediate child of BroadcastHashJoinExec
+            val parent = plan.find(_.children.contains(columnarToRowExec.head))
+            assert(parent.get.isInstanceOf[BroadcastHashJoinExec])
+
+            // There should be a CometBroadcastExchangeExec under CometColumnarToRowExec
+            val broadcastQueryStage =
+              columnarToRowExec.head.find(_.isInstanceOf[BroadcastQueryStageExec])
+            assert(broadcastQueryStage.isDefined)
+            assert(
+              broadcastQueryStage.get
+                .asInstanceOf[BroadcastQueryStageExec]
+                .broadcast
+                .isInstanceOf[CometBroadcastExchangeExec])
+          }
+        }
+      }
+    }
+  }
+
   test("expand operator") {
     val data1 = (0 until 1000)
       .map(_ % 5) // reduce value space to trigger dictionary encoding
@@ -821,11 +936,6 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("explain native plan") {
-    // https://github.com/apache/datafusion-comet/issues/1441
-    assume(!CometConf.isExperimentalNativeScan)
-    // there are no assertions in this test to prove that the explain feature
-    // wrote the expected output to stdout, but we at least test that enabling
-    // the config does not cause any exceptions.
     withSQLConf(
       CometConf.COMET_EXPLAIN_NATIVE_ENABLED.key -> "true",
       SQLConf.AUTO_BROADCASTJOIN_THRESHOLD.key -> "-1") {
@@ -1390,6 +1500,10 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("bucketed table") {
+    // native_datafusion actually passes this test, but in the case where buckets are pruned it fails, so we're
+    // falling back for bucketed scans entirely as a workaround.
+    // https://github.com/apache/datafusion-comet/issues/1719
+    assume(CometConf.COMET_NATIVE_SCAN_IMPL.get() != CometConf.SCAN_NATIVE_DATAFUSION)
     val bucketSpec = Some(BucketSpec(8, Seq("i", "j"), Nil))
     val bucketedTableTestSpecLeft = BucketedTableTestSpec(bucketSpec, expectedShuffle = false)
     val bucketedTableTestSpecRight = BucketedTableTestSpec(bucketSpec, expectedShuffle = false)
@@ -1585,7 +1699,6 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("Fallback to Spark for TakeOrderedAndProjectExec with offset") {
-    assume(isSpark34Plus)
     Seq("true", "false").foreach(aqeEnabled =>
       withSQLConf(SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> aqeEnabled) {
         withTable("t1") {
@@ -1702,13 +1815,11 @@ class CometExecSuite extends CometTestBase {
 
                 // Verify that the BatchScanExec nodes supported columnar output when requested for Spark 3.4+.
                 // Earlier versions support columnar output for fewer type.
-                if (isSpark34Plus) {
-                  val leaves = df.queryExecution.executedPlan.collectLeaves()
-                  if (parquetVectorized && isSpark34Plus) {
-                    assert(leaves.forall(_.supportsColumnar))
-                  } else {
-                    assert(!leaves.forall(_.supportsColumnar))
-                  }
+                val leaves = df.queryExecution.executedPlan.collectLeaves()
+                if (parquetVectorized) {
+                  assert(leaves.forall(_.supportsColumnar))
+                } else {
+                  assert(!leaves.forall(_.supportsColumnar))
                 }
               }
             }
@@ -1748,6 +1859,8 @@ class CometExecSuite extends CometTestBase {
   }
 
   test("SparkToColumnar eliminate redundant in AQE") {
+    // TODO fix for Spark 4.0.0
+    assume(!isSpark40Plus)
     withSQLConf(
       SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> "true",
       CometConf.COMET_SHUFFLE_MODE.key -> "jvm") {
@@ -1916,6 +2029,59 @@ class CometExecSuite extends CometTestBase {
 
     assert(!CometScanExec.isFileFormatSupported(new CustomParquetFileFormat()))
   }
+
+  test("SparkToColumnar override node name for row input") {
+    // TODO fix for Spark 4.0.0
+    assume(!isSpark40Plus)
+    withSQLConf(
+      SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> "true",
+      CometConf.COMET_SHUFFLE_MODE.key -> "jvm") {
+      val df = spark
+        .range(1000)
+        .selectExpr("id as key", "id % 8 as value")
+        .toDF("key", "value")
+        .groupBy("key")
+        .count()
+      df.collect()
+
+      val planAfter = df.queryExecution.executedPlan
+      assert(planAfter.toString.startsWith("AdaptiveSparkPlan isFinalPlan=true"))
+      val adaptivePlan = planAfter.asInstanceOf[AdaptiveSparkPlanExec].executedPlan
+      val nodeNames = adaptivePlan.collect { case c: CometSparkToColumnarExec =>
+        c.nodeName
+      }
+      assert(nodeNames.length == 1)
+      assert(nodeNames.head == "CometSparkRowToColumnar")
+    }
+  }
+
+  test("SparkToColumnar override node name for columnar input") {
+    withSQLConf(
+      SQLConf.USE_V1_SOURCE_LIST.key -> "",
+      SQLConf.ADAPTIVE_EXECUTION_ENABLED.key -> "false",
+      CometConf.COMET_NATIVE_SCAN_ENABLED.key -> "false",
+      CometConf.COMET_CONVERT_FROM_PARQUET_ENABLED.key -> "true") {
+      withTempDir { dir =>
+        var df = spark
+          .range(10000)
+          .selectExpr("id as key", "id % 8 as value")
+          .toDF("key", "value")
+
+        df.write.mode("overwrite").parquet(dir.toString)
+        df = spark.read.parquet(dir.toString)
+        df = df.groupBy("key", "value").count()
+        df.collect()
+
+        val planAfter = df.queryExecution.executedPlan
+        val nodeNames = planAfter.collect { case c: CometSparkToColumnarExec =>
+          c.nodeName
+        }
+        assert(nodeNames.length == 1)
+        assert(nodeNames.head == "CometSparkColumnarToColumnar")
+      }
+    }
+  }
+
 }
 
 case class BucketedTableTestSpec(
